/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#include "message_types.h"
#include "motionSL_hash.h"
#include <mcx-cpp/message_types.h>
#include <mcx-cpp/pb_decode.h>

namespace mcx_robot_control {

template <typename MSG_TYPE>
static uint8_t* fillHeader(nng_msg* nng_msg) {
  const auto cursor = static_cast<uint8_t*>(nng_msg_body(nng_msg));
  *reinterpret_cast<uint32_t*>(cursor) = motion_spec::getHash<MSG_TYPE>();
  return cursor + motion_spec::getHashSize();
}

static bool encodeMotionCommandHook(pb_ostream_t* stream, const pb_field_t* field, void* const* arg) {
  const std::shared_ptr<MotionCommand> motion_cmd = *static_cast<std::shared_ptr<MotionCommand>*>(*arg);

  if (!pb_encode_tag_for_field(stream, field)) {
    return false;
  }

  if (stream->state) {
    return motion_cmd->encode(stream) > 0;
  }

  stream->bytes_written += motion_cmd->size();
  return true;
}

static bool encodeMotionContainerHook(pb_ostream_t* stream, const pb_field_t* field, void* const* arg) {
  const std::vector<std::shared_ptr<MotionCommand>>* motion_cmd =
      static_cast<std::vector<std::shared_ptr<MotionCommand>>*>(*arg);

  for (auto cmd : *motion_cmd) {
    motion_spec_MotionCommand mc{};
    mc.id = static_cast<uint32_t>(cmd->id());
    mc.commandtype = cmd->type();
    mc.commandarguments.funcs.encode = encodeMotionCommandHook;
    mc.commandarguments.arg = &cmd;
    if (!pb_encode_tag_for_field(stream, field)) {
      return false;
    }
    if (!pb_encode_submessage(stream, motion_spec_MotionCommand_fields, &mc)) {
      return false;
    }
  }

  return true;
}

nng_msg* encodeMotionProgram(std::vector<std::shared_ptr<MotionCommand>> motion_commands, int id,
                             const std::string& name) {

  motion_spec_MotionProgram motion_program{};
  motion_program.id = static_cast<uint32_t>(id);
  name.copy(motion_program.name, sizeof(motion_program));
  motion_program.commandlist.arg = &motion_commands;
  motion_program.commandlist.funcs.encode = &encodeMotionContainerHook;

  // measure required message size;
  pb_ostream_t stream = PB_OSTREAM_SIZING;
  pb_encode(&stream, motion_spec_MotionProgram_fields, &motion_program);
  const auto estimated_size = stream.bytes_written;

  const auto msg = mcx_cpp::allocateMsg(stream.bytes_written);
  const auto cursor = fillHeader<motion_spec_MotionProgram>(msg);
  stream = pb_ostream_from_buffer(cursor, stream.bytes_written);
  if (pb_encode(&stream, motion_spec_MotionProgram_fields, &motion_program)) {
    nng_msg_chop(msg, mcx_cpp::choppedSize(estimated_size, stream.bytes_written));
    return msg;
  }
  nng_msg_free(msg);

  return nullptr;
}

nng_msg* encodeJointToCart(const motion_spec_JointToCart& joint_to_cart) {
  const auto msg = mcx_cpp::allocateMsg(motion_spec_JointToCart_size);
  const auto cursor = fillHeader<motion_spec_JointToCart>(msg);
  auto stream = pb_ostream_from_buffer(cursor, motion_spec_JointToCart_size);
  if (pb_encode(&stream, motion_spec_JointToCart_fields, &joint_to_cart)) {
    nng_msg_chop(msg, mcx_cpp::choppedSize(motion_spec_JointToCart_size, stream.bytes_written));
    return msg;
  }
  nng_msg_free(msg);
  return nullptr;
}

nng_msg* encodeCartToJoint(const motion_spec_CartToJoint& cart_to_joint) {
  const auto msg = mcx_cpp::allocateMsg(motion_spec_CartToJoint_size);
  const auto cursor = fillHeader<motion_spec_CartToJoint>(msg);
  auto stream = pb_ostream_from_buffer(cursor, motion_spec_CartToJoint_size);
  if (pb_encode(&stream, motion_spec_CartToJoint_fields, &cart_to_joint)) {
    nng_msg_chop(msg, mcx_cpp::choppedSize(motion_spec_CartToJoint_size, stream.bytes_written));
    return msg;
  }
  nng_msg_free(msg);
  return nullptr;
}

JointToCart decodeJointToCart(nng_msg* nng_msg) {
  motion_spec_JointToCart joint_to_cart{};
  size_t size;
  const auto msg = mcx_cpp::getMsg(nng_msg, size);
  pb_istream_t stream = pb_istream_from_buffer(msg, size);
  if (pb_decode(&stream, motion_spec_JointToCart_fields, &joint_to_cart)) {
    return JointToCart{mcx_cpp::McxStatus{mcx_cpp::StatusCode::OK, mcx_cpp::McxStatus::NO_ERROR}, joint_to_cart};
  }

  return JointToCart{
      mcx_cpp::McxStatus{mcx_cpp::StatusCode::FAILED, mcx_cpp::McxStatus::CLIENT_FAILED_TO_DECODE_REPLY}};
}

CartToJoint decodeCartToJoint(nng_msg* nng_msg) {
  motion_spec_CartToJoint cart_to_joint{};
  size_t size;
  const auto msg = mcx_cpp::getMsg(nng_msg, size);
  pb_istream_t stream = pb_istream_from_buffer(msg, size);
  if (pb_decode(&stream, motion_spec_CartToJoint_fields, &cart_to_joint)) {
    return CartToJoint{mcx_cpp::McxStatus{mcx_cpp::StatusCode::OK, mcx_cpp::McxStatus::NO_ERROR}, cart_to_joint};
  }

  return CartToJoint{
      mcx_cpp::McxStatus{mcx_cpp::StatusCode::FAILED, mcx_cpp::McxStatus::CLIENT_FAILED_TO_DECODE_REPLY}};
}

} // namespace mcx_robot_control