/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#include "movel.h"

namespace mcx_robot_control {

MoveL::MoveL(size_t id, std::vector<Waypoint> waypoints, double velocity, double acceleration,
             double rotational_velocity, double rotational_acceleration, std::vector<double> ref_joint_coord_rad)
    : container_{.id = id,
                 .waypoints{std::move(waypoints)},
                 .velocity = velocity,
                 .acceleration = acceleration,
                 .rotational_velocity = rotational_velocity,
                 .rotational_acceleration = rotational_acceleration,
                 .ref_joint_coord_rad{std::move(ref_joint_coord_rad)}} {

  // measure message size;
  pb_ostream_t stream = PB_OSTREAM_SIZING;
  size_ = encode(&stream);
}

bool MoveL::encodeWaypointsHook(pb_ostream_t* stream, const pb_field_t* field, void* const* arg) {
  const auto container_ptr = static_cast<Container*>(*arg);
  for (auto& w : container_ptr->waypoints) {
    motion_spec_CartWaypoint cw{};
    cw.has_constraint = true;
    cw.constraint = {.type = w.type(), .factor = w.smoothFactor()};
    cw.has_segmentVelocity = true;
    cw.segmentVelocity = w.segmentVelocity();
    size_t i = 0;
    for (auto& coord : cw.pose.coordinates) {
      coord = w.coordinate(i++);
    }
    cw.pose.coordinates_count = static_cast<pb_size_t>(w.length());
    if (!pb_encode_tag_for_field(stream, field)) {
      return false;
    }
    if (!pb_encode_submessage(stream, motion_spec_CartWaypoint_fields, &cw)) {
      return false;
    }
  }
  return true;
}

size_t MoveL::encode(pb_ostream_t* stream) const {
  motion_spec_MoveL move_l{};
  Container data{container_};
  auto& constraint = move_l.constraint;
  constraint.type = motion_spec_MOTIONCONSTRAINTTYPE_VELANDACC;
  constraint.has_velacc_values = true;
  auto& velacc = constraint.velacc_values;
  velacc.vMax = container_.velocity;
  velacc.aMax = container_.acceleration;
  velacc.has_omegaMax = true;
  velacc.omegaMax = container_.rotational_velocity;
  velacc.has_alfaMax = true;
  velacc.alfaMax = container_.rotational_acceleration;

  auto& number_of_coords = move_l.referenceJoint.coordinates_count;
  for (auto& c : move_l.referenceJoint.coordinates) {
    if (number_of_coords < container_.ref_joint_coord_rad.size()) {
      c = container_.ref_joint_coord_rad[number_of_coords++];
    }
  }

  move_l.waypoints.arg = &data;
  move_l.waypoints.funcs.encode = &MoveL::encodeWaypointsHook;

  if (pb_encode_submessage(stream, motion_spec_MoveL_fields, &move_l)) {
    return stream->bytes_written;
  }
  return 0;
}

size_t MoveL::size() const { return size_; }

motion_spec_MOTIONTYPE MoveL::type() const { return motion_spec_MOTIONTYPE_CARTMOTION; }

size_t MoveL::id() const { return container_.id; }

} // namespace mcx_robot_control