/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#include "motion_program.h"
#include "message_types.h"
#include "movec.h"
#include "movej.h"
#include "movel.h"
#include "wait.h"

namespace mcx_robot_control {

MotionProgram::MotionProgram(const mcx_cpp::Request& request) : request_{request} {}

void MotionProgram::clear() {
  motion_commands_.clear();
  cmd_counter_ = DEFAULT_COUNTER;
}

size_t MotionProgram::addMoveC(std::vector<Waypoint> waypoints, double angle, double velocity, double acceleration,
                               double rotational_velocity, double rotational_acceleration,
                               std::vector<double> ref_joint_coord_rad) {
  std::shared_ptr<MotionCommand> cmd =
      std::make_shared<MoveC>(MoveC{cmd_counter_++, std::move(waypoints), angle, velocity, acceleration,
                                    rotational_velocity, rotational_acceleration, std::move(ref_joint_coord_rad)});
  return emplaceBack(motion_commands_, std::move(cmd))->id();
}

size_t MotionProgram::addMoveL(Waypoint waypoint, double velocity, double acceleration, double rotational_velocity,
                               double rotational_acceleration, std::vector<double> ref_joint_coord_rad) {
  return addMoveL(std::vector{std::move(waypoint)}, velocity, acceleration, rotational_velocity,
                  rotational_acceleration, std::move(ref_joint_coord_rad));
}

size_t MotionProgram::addMoveL(std::vector<Waypoint> waypoints, double velocity, double acceleration,
                               double rotational_velocity, double rotational_acceleration,
                               std::vector<double> ref_joint_coord_rad) {
  std::shared_ptr<MotionCommand> cmd = std::make_shared<MoveL>(MoveL{cmd_counter_++,
                                                                     {std::move(waypoints)},
                                                                     velocity,
                                                                     acceleration,
                                                                     rotational_velocity,
                                                                     rotational_acceleration,
                                                                     std::move(ref_joint_coord_rad)});
  return emplaceBack(motion_commands_, std::move(cmd))->id();
}

size_t MotionProgram::addMoveJ(Waypoint waypoint, double rotational_velocity, double rotational_acceleration) {
  return addMoveJ(std::vector{std::move(waypoint)}, rotational_velocity, rotational_acceleration);
}

size_t MotionProgram::addMoveJ(std::vector<Waypoint> waypoints, double rotational_velocity,
                               double rotational_acceleration) {
  std::shared_ptr<MotionCommand> cmd = std::make_shared<MoveJ>(
      MoveJ{cmd_counter_++, std::move(waypoints), rotational_velocity, rotational_acceleration});
  return emplaceBack(motion_commands_, std::move(cmd))->id();
}

size_t MotionProgram::addWait(double timeout) {
  std::shared_ptr<MotionCommand> cmd = std::make_shared<Wait>(Wait{cmd_counter_++, timeout, "", false});
  return emplaceBack(motion_commands_, std::move(cmd))->id();
}

size_t MotionProgram::addWait(const std::string& path, bool equal_to, double timeout) {
  std::shared_ptr<MotionCommand> cmd = std::make_shared<Wait>(Wait{cmd_counter_++, timeout, path, equal_to});
  return emplaceBack(motion_commands_, std::move(cmd))->id();
}

mcx_cpp::StatusReply MotionProgram::send(const std::string& name, int id) const {
  return std::async(
      [](auto worker, std::vector<std::shared_ptr<MotionCommand>> motion_commands, int id, const std::string& name,
         unsigned int timeout_ms) {
        worker.setMsg(encodeMotionProgram(std::move(motion_commands), id, name));
        if (worker.msg()) {
          auto status = mcx_cpp::Request::send(worker, timeout_ms);
          if (status.status() == mcx_cpp::StatusCode::OK) {
            status = mcx_cpp::McxStatus{mcx_cpp::decodeStatusMsg(worker.msg())};
            return status;
          }
          return mcx_cpp::McxStatus{status};
        }

        return mcx_cpp::McxStatus{mcx_cpp::StatusCode::FAILED, mcx_cpp::McxStatus::CLIENT_FAILED_TO_ENCODE_REQUEST};
      },
      request_.getWorker(), motion_commands_, id, name, request_.getConnectionOptions().io_timeout_ms);
}

} // namespace mcx_robot_control