/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#ifndef MOTORCORTEX_CPP_MOTION_PROGRAM_H
#define MOTORCORTEX_CPP_MOTION_PROGRAM_H

#include "motion_command.h"
#include "waypoint.h"
#include <mcx-cpp/request.h>

namespace mcx_robot_control {

/**@brief Class represents a motion program of the manipulator.
 * @addtogroup motion program
 * @{
 */

class MotionProgram {
  template <class C, class... Args>
  auto emplaceBack(C& c, Args&&... args) -> decltype(c.back()) {
    c.emplace_back(std::forward<Args>(args)...);
    return c.back();
  }

  static constexpr size_t DEFAULT_COUNTER = 1;

public:
  /**@brief Creates new Motion Program.
   * @param request(motorcortex.Request): reference to a Request instance
   */
  explicit MotionProgram(const mcx_cpp::Request& request);

  /**@brief Clears all commands in the program
   */
  void clear();

  /**@brief Adds a MoveC(circular move) command to the program
   * @param waypoints: a list of waypoints
   * @param angle: rotation angle, rad
   * @param velocity: maximum velocity, m/sec
   * @param acceleration: maximum acceleration, m/sec^2
   * @param rotational_velocity: maximum joint velocity, rad/sec
   * @param rotational_acceleration: maximum joint acceleration, rad/sec^2
   * @param ref_joint_coord_rad: reference joint coordinates for the first waypoint
   * @return id of the command
   */
  size_t addMoveC(std::vector<Waypoint> waypoints, double angle = 0, double velocity = 0.1, double acceleration = 0.2,
                  double rotational_velocity = 3.18, double rotational_acceleration = 6.37,
                  std::vector<double> ref_joint_coord_rad = {});

  /**@brief Adds a MoveL(Linear move) command to the program
   * @param waypoint: a single waypoint
   * @param velocity: maximum velocity, m/sec
   * @param acceleration: maximum acceleration, m/sec^2
   * @param rotational_velocity: maximum joint velocity, rad/sec
   * @param rotational_acceleration: maximum joint acceleration, rad/sec^2
   * @param ref_joint_coord_rad: reference joint coordinates for the first waypoint
   * @return id of the command
   */
  size_t addMoveL(Waypoint waypoint, double velocity = 0.1, double acceleration = 0.2,
                  double rotational_velocity = 3.18, double rotational_acceleration = 6.37,
                  std::vector<double> ref_joint_coord_rad = {});

  /**@brief Adds a MoveL(Linear move) command to the program
   * @param waypoints: a list of waypoints
   * @param velocity: maximum velocity, m/sec
   * @param acceleration: maximum acceleration, m/sec^2
   * @param rotational_velocity: maximum joint velocity, rad/sec
   * @param rotational_acceleration: maximum joint acceleration, rad/sec^2
   * @param ref_joint_coord_rad: reference joint coordinates for the first waypoint
   * @return id of the command
   */
  size_t addMoveL(std::vector<Waypoint> waypoints, double velocity = 0.1, double acceleration = 0.2,
                  double rotational_velocity = 3.18, double rotational_acceleration = 6.37,
                  std::vector<double> ref_joint_coord_rad = {});

  /**@brief Adds MoveJ(Joint move) command to the program
   * @param waypoint: a single waypoint
   * @param rotational_velocity: maximum joint velocity, rad/sec
   * @param rotational_acceleration: maximum joint acceleration, rad/sec^2
   * @return id of the command
   */
  size_t addMoveJ(Waypoint waypoint, double rotational_velocity = 3.18, double rotational_acceleration = 6.37);

  /**@brief Adds MoveJ(Joint move) command to the program
   * @param waypoints: a list of waypoints
   * @param rotational_velocity: maximum joint velocity, rad/sec
   * @param rotational_acceleration: maximum joint acceleration, rad/sec^2
   * @return id of the command
   */
  size_t addMoveJ(std::vector<Waypoint> waypoints, double rotational_velocity = 3.18,
                  double rotational_acceleration = 6.37);

  /**@brief Adds Wait command to the program.
   * @param timeout: timeout in seconds
   * @return id of the command
   */
  size_t addWait(double timeout);

  /**@brief Adds Wait command to the program.
   * @param path: path to the parameter that will be compared to value
   * @param equal_to: value that the parameter is compared to
   * @param timeout: timeout in seconds
   * @return id of the command
   */
  size_t addWait(const std::string& path, bool equal_to, double timeout = std::numeric_limits<double>::max());

  /**@brief Sends program to a robot
   * @param program_name: program name
   * @param id: program id
   * @return status of the send operation
   */
  mcx_cpp::StatusReply send(const std::string& name, int id) const;

  /** @} End of Doxygen Groups*/

private:
  size_t cmd_counter_{DEFAULT_COUNTER};
  const mcx_cpp::Request& request_;
  std::vector<std::shared_ptr<MotionCommand>> motion_commands_;
};

} // namespace mcx_robot_control

#endif // MOTORCORTEX_CPP_MOTION_PROGRAM_H
