/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#include "pose_transformer.h"
#include "message_types.h"

namespace mcx_robot_control {

PoseTransformer::PoseTransformer(const mcx_cpp::Request& request) : request_{request} {}

JointToCartReply PoseTransformer::calcJointToCartPose(const JointToCart& joint_pose) const {
  return std::async(
      [](auto worker, const JointToCart& joint_pose, unsigned int timeout_ms) {
        worker.setMsg(encodeJointToCart(joint_pose.joint_to_cart_));
        if (worker.msg()) {
          auto status = mcx_cpp::Request::send(worker, timeout_ms);
          if (status.status() == mcx_cpp::StatusCode::OK) {
            auto reply{decodeJointToCart(worker.msg())};
            return reply;
          }
          return JointToCart{status};
        }

        return JointToCart{
            mcx_cpp::McxStatus{mcx_cpp::StatusCode::FAILED, mcx_cpp::McxStatus::CLIENT_FAILED_TO_ENCODE_REQUEST}};
      },
      request_.getWorker(), joint_pose, request_.getConnectionOptions().io_timeout_ms);
}

JointToCartReply PoseTransformer::calcJointToCartPose(const std::vector<double>& joint_pose) const {
  return calcJointToCartPose(joint_pose, {});
}

JointToCartReply PoseTransformer::calcJointToCartPose(const CartToJoint& joint_pose) const {
  return calcJointToCartPose(
      JointToCart{mcx_cpp::McxStatus{joint_pose.status(), static_cast<JointToCart::ErrorCode>(joint_pose.error())},
                  motion_spec_JointToCart{.cartpose = joint_pose.cartPose(),
                                          .jointpose = joint_pose.jointPose(),
                                          .jointtwist = joint_pose.jointVelocity(),
                                          .carttwist = joint_pose.cartTwist()}});
}

JointToCartReply PoseTransformer::calcJointToCartPose(const std::vector<double>& joint_pose,
                                                      const std::vector<double>& actual_cart_pose) const {
  motion_spec_JointToCart joint_to_cart{};
  auto& jp = joint_to_cart.jointpose;
  jp.has_id = true;

  static constexpr size_t MAX_JOINT_NUMBER = sizeof(jp.coordinates) / sizeof(jp.coordinates[0]);
  for (auto& v : joint_pose) {
    if (jp.coordinates_count < MAX_JOINT_NUMBER) {
      jp.coordinates[jp.coordinates_count] = v;
    }
    ++jp.coordinates_count;
  }
  joint_to_cart.jointtwist.coordinates_count = jp.coordinates_count;

  auto& cp = joint_to_cart.cartpose;
  static constexpr size_t MAX_CART_NUMBER = sizeof(cp.coordinates) / sizeof(cp.coordinates[0]);
  for (auto& v : actual_cart_pose) {
    if (cp.coordinates_count < MAX_CART_NUMBER) {
      cp.coordinates[cp.coordinates_count] = v;
    }
    ++cp.coordinates_count;
  }
  joint_to_cart.carttwist.coordinates_count = cp.coordinates_count;

  return calcJointToCartPose(
      JointToCart{mcx_cpp::McxStatus{mcx_cpp::StatusCode::OK, mcx_cpp::McxStatus::ErrorCode::NO_ERROR}, joint_to_cart});
}

CartToJointReply PoseTransformer::calcCartToJointPose(const CartToJoint& cart_pose) const {
  return std::async(
      [](auto worker, const CartToJoint& cart_pose, unsigned int timeout_ms) {
        worker.setMsg(encodeCartToJoint(cart_pose.cart_to_joint_));
        if (worker.msg()) {
          auto status = mcx_cpp::Request::send(worker, timeout_ms);
          if (status.status() == mcx_cpp::StatusCode::OK) {
            auto reply{decodeCartToJoint(worker.msg())};
            return reply;
          }
          return CartToJoint{status};
        }

        return CartToJoint{
            mcx_cpp::McxStatus{mcx_cpp::StatusCode::FAILED, mcx_cpp::McxStatus::CLIENT_FAILED_TO_ENCODE_REQUEST}};
      },
      request_.getWorker(), cart_pose, request_.getConnectionOptions().io_timeout_ms);
}

CartToJointReply PoseTransformer::calcCartToJointPose(const std::vector<double>& cart_pose,
                                                      motion_spec_FEATUREFRAMETYPE frame_type) const {
  return calcCartToJointPose(cart_pose, {}, frame_type);
}

CartToJointReply PoseTransformer::calcCartToJointPose(const std::vector<double>& cart_pose,
                                                      const std::vector<double>& ref_joint_pose,
                                                      motion_spec_FEATUREFRAMETYPE frame_type) const {
  motion_spec_CartToJoint cart_to_joint{};
  cart_to_joint.frame_type = frame_type;
  auto& jp = cart_to_joint.jointpose;
  jp.has_id = true;
  static constexpr size_t MAX_JOINT_NUMBER = sizeof(jp.coordinates) / sizeof(jp.coordinates[0]);
  for (auto& v : ref_joint_pose) {
    if (jp.coordinates_count < MAX_JOINT_NUMBER) {
      jp.coordinates[jp.coordinates_count] = v;
    }
    ++jp.coordinates_count;
  }
  if (jp.coordinates_count == 0) {
    jp.coordinates_count = 6;
  }
  cart_to_joint.jointtwist.coordinates_count = jp.coordinates_count;

  auto& cp = cart_to_joint.cartpose;
  static constexpr size_t MAX_CART_NUMBER = sizeof(cp.coordinates) / sizeof(cp.coordinates[0]);
  for (auto& v : cart_pose) {
    if (cp.coordinates_count < MAX_CART_NUMBER) {
      cp.coordinates[cp.coordinates_count] = v;
    }
    ++cp.coordinates_count;
  }
  cart_to_joint.carttwist.coordinates_count = cp.coordinates_count;

  return calcCartToJointPose(
      CartToJoint{mcx_cpp::McxStatus{mcx_cpp::StatusCode::FAILED, mcx_cpp::McxStatus::CLIENT_FAILED_TO_ENCODE_REQUEST},
                  cart_to_joint});
}

CartToJointReply PoseTransformer::calcCartToJointPose(const JointToCart& cart_pose,
                                                      motion_spec_FEATUREFRAMETYPE frame_type) const {
  return calcCartToJointPose(
      CartToJoint{mcx_cpp::McxStatus{cart_pose.status(), static_cast<CartToJoint::ErrorCode>(cart_pose.error())},
                  motion_spec_CartToJoint{.cartpose = cart_pose.cartPose(),
                                          .jointpose = cart_pose.jointPose(),
                                          .frame_type = frame_type,
                                          .jointtwist = cart_pose.jointVelocity(),
                                          .carttwist = cart_pose.cartTwist()}});
}

} // namespace mcx_robot_control