/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#include "robot_command.h"
#include "utils.h"

#include <mcx-cpp/mcx-cpp.h>

namespace mcx_robot_control {

RobotCommand::RobotCommand(const mcx_cpp::Request& request)
    : request_{request}, paths_{
                             R"({"LOGIC_STATE_COMMAND": "root/Logic/stateCommand",
  "LOGIC_MODE_COMMAND": "root/Logic/modeCommand",
  "INTERPRETER_COMMAND": "root/MotionInterpreter/interpreterCommand",
  "ACTIVATE_SEMIAUTO": "root/ManipulatorControl/activateSemiAuto",
  "ACTIVATE_MOVE_TO_START": "root/ManipulatorControl/activateMoveToStart",
  "LOGIC_STATE": "root/Logic/state",
  "LOGIC_MODE": "root/Logic/mode",
  "INTERPRETER_STATE": "root/MotionInterpreter/actualStateOut",
  "SEMIAUTO_STATE": "root/ManipulatorControl/semiAutoMotionGenerator/motionGenStateOut",
  "MOVE_TO_START_STATE": "root/ManipulatorControl/moveToStartMotionGenerator/motionGenStateOut",
  "SEMIAUTO_TARGET": "root/ManipulatorControl/semiAutoMotionGenerator/target",
  "SEMIAUTO_MAX_VEL": "root/ManipulatorControl/semiAutoMotionGenerator/maxVel",
  "SEMIAUTO_MAX_ACC": "root/ManipulatorControl/semiAutoMotionGenerator/maxAcc"})"_json} {}

CommandReply RobotCommand::acknowledge(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        RobotStatus status{mcx_cpp::StatusCode::OK, RobotStatus::NO_ERROR};

        States entry_state{};
        timeout_sec -= measureSec([&] {
          // checks active state
          req.getParameter(paths.get("LOGIC_STATE")).get().value(entry_state);
          // sends acknowledge command
          req.setParameter(paths.get("LOGIC_STATE_COMMAND"), StateEvents::ACKNOWLEDGE_ERROR);
          // waits until the system leaves E-Stop reset
        });

        States robot_state{entry_state};
        do {
          // decrementing timer
          timeout_sec -= measureSec([&] { req.getParameter(paths.get("LOGIC_STATE")).get().value(robot_state); });
          if (timeout_sec < 0) {
            const auto src_state = std::to_string(static_cast<std::underlying_type_t<States>>(entry_state));
            const auto cur_state = std::to_string(static_cast<std::underlying_type_t<States>>(robot_state));

            status = {mcx_cpp::StatusCode::FAILED, RobotStatus::SET_LOGIC_STATE_TIMEOUT,
                      std::string("Timeout, can not acknowledge from state: ")
                          .append(src_state)
                          .append(", state at timeout: ")
                          .append(cur_state)};
            break;
          }
        } while (robot_state == States::ESTOP_OFF_S || robot_state == States::RESET_ESTOP_T ||
                 robot_state == States::FORCEDISENGAGED_S);

        return status;
      },
      std::ref(request_), timeout_sec, paths_);
}

// state commands
CommandReply RobotCommand::off(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        timeout_sec -= measureSec([&] { req.setParameter(paths.get("LOGIC_STATE_COMMAND"), StateEvents::GOTO_OFF_E); });
        return switchToLogicState(req, timeout_sec, States::OFF_S, "Off state", paths).first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::disengage(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        timeout_sec -=
            measureSec([&] { req.setParameter(paths.get("LOGIC_STATE_COMMAND"), StateEvents::GOTO_DISENGAGED_E); });
        return switchToLogicState(req, timeout_sec, States::DISENGAGED_S, "Disengaged state", paths).first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::engage(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        timeout_sec -=
            measureSec([&] { req.setParameter(paths.get("LOGIC_STATE_COMMAND"), StateEvents::GOTO_ENGAGED_E); });
        return switchToLogicState(req, timeout_sec, States::ENGAGED_S, "Engaged state", paths).first;
      },
      std::ref(request_), timeout_sec, paths_);
}

// mode commands
CommandReply RobotCommand::pauseMode(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        // sets system to manual cart
        timeout_sec -= measureSec([&] { req.setParameter(paths.get("LOGIC_MODE_COMMAND"), ModeEvents::GOTO_PAUSE_E); });
        // waits for the mode
        return switchToLogicMode(req, timeout_sec, Modes::PAUSE_M, "Pause mode", paths).first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::manualCartMode(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        timeout_sec -=
            measureSec([&] { req.setParameter(paths.get("LOGIC_MODE_COMMAND"), ModeEvents::GOTO_MANUAL_CART_MODE_E); });
        return switchToLogicMode(req, timeout_sec, Modes::MANUAL_CART_MODE_M, "Manual cartesian mode", paths).first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::manualJointMode(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        timeout_sec -= measureSec(
            [&] { req.setParameter(paths.get("LOGIC_MODE_COMMAND"), ModeEvents::GOTO_MANUAL_JOINT_MODE_E); });
        return switchToLogicMode(req, timeout_sec, Modes::MANUAL_JOINT_MODE_M, "Manual joint mode", paths).first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::autoMode(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        timeout_sec -=
            measureSec([&] { req.setParameter(paths.get("LOGIC_MODE_COMMAND"), ModeEvents::GOTO_AUTO_RUN_E); });
        return switchToLogicMode(req, timeout_sec, Modes::AUTO_RUN_M, "Auto-run mode", paths).first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::semiAutoMode(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        timeout_sec -=
            measureSec([&] { req.setParameter(paths.get("LOGIC_MODE_COMMAND"), ModeEvents::GOTO_SEMI_AUTO_E); });
        // waits for the mode
        return switchToLogicMode(req, timeout_sec, Modes::SEMI_AUTO_M, "Semi-Auto mode", paths).first;
      },
      std::ref(request_), timeout_sec, paths_);
}

// guard move commands
CommandReply RobotCommand::moveToPoint(const std::vector<double>& angles_rad, double max_vel, double max_acc,
                                       double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, const std::vector<double>& angles_rad, const double max_vel, const double max_acc,
         double timeout_sec, const Paths& paths) {
        timeout_sec -=
            measureSec([&] { req.setParameter(paths.get("LOGIC_MODE_COMMAND"), ModeEvents::GOTO_SEMI_AUTO_E); });
        // waits for the mode
        auto semi_auto_mode = switchToLogicMode(req, timeout_sec, Modes::SEMI_AUTO_M, "Semi-Auto mode", paths);
        if (semi_auto_mode.first.status() != mcx_cpp::StatusCode::OK) {
          return semi_auto_mode.first;
        }
        req.setParameter(paths.get("SEMIAUTO_MAX_VEL"), max_vel).get();
        req.setParameter(paths.get("SEMIAUTO_MAX_ACC"), max_acc).get();
        req.setParameter(paths.get("SEMIAUTO_TARGET"), angles_rad).get();
        MotionGeneratorStates semi_auto_motion_generator_state{};
        auto mode{Modes::INIT_M};
        do {
          timeout_sec -= measureSec([&] {
            // activates guard
            req.setParameter(paths.get("ACTIVATE_SEMIAUTO"), true).get();
            constexpr size_t SHORT_SLEEP_MSEC = 1;
            std::this_thread::sleep_for(std::chrono::milliseconds(SHORT_SLEEP_MSEC));
            req.getParameter(paths.get("SEMIAUTO_STATE")).get().value(semi_auto_motion_generator_state);
            req.getParameter(paths.get("LOGIC_MODE")).get().value(mode);
          });
          if (timeout_sec < 0) {
            return RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::MOVE_TIMEOUT,
                               "Timeout, can not complete move to start"};
          }
        } while (semi_auto_motion_generator_state == MotionGeneratorStates::RUNNING_S && mode == Modes::SEMI_AUTO_M);
        return semi_auto_mode.first;
      },
      std::ref(request_), angles_rad, max_vel, max_acc, timeout_sec, paths_);
}

CommandReply RobotCommand::moveToStart(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        // If interpreter in pause, moveToStart does nothing
        InterpreterStates interpreter_state{};
        req.getParameter(paths.get("INTERPRETER_STATE")).get().value(interpreter_state);
        if (interpreter_state == InterpreterStates::PROGRAM_PAUSE_S) {
          return RobotStatus{mcx_cpp::StatusCode::OK, RobotStatus::NO_ERROR};
        }

        // checks if program is playing
        auto motion_not_allowed = switchToInterpreterState(
            req, timeout_sec, InterpreterStates::MOTION_NOT_ALLOWED_S,
            {InterpreterStates::PROGRAM_STOP_S, InterpreterStates::MSL_PROGRAM_DONE_S}, "Move to Start state", paths);
        if (motion_not_allowed.second == InterpreterStates::MOTION_NOT_ALLOWED_S) {
          // commands move to start
          req.setParameter(paths.get("INTERPRETER_COMMAND"), InterpreterEvents::MOVE_TO_START);

          MotionGeneratorStates semi_auto_state{};
          do {
            timeout_sec -= measureSec([&] {
              // activates guard
              req.setParameter(paths.get("ACTIVATE_MOVE_TO_START"), true).get();
              constexpr size_t SHORT_SLEEP_MSEC = 1;
              std::this_thread::sleep_for(std::chrono::milliseconds(SHORT_SLEEP_MSEC));
              req.getParameter(paths.get("MOVE_TO_START_STATE")).get().value(semi_auto_state);
              req.getParameter(paths.get("INTERPRETER_STATE")).get().value(interpreter_state);
            });
            if (timeout_sec < 0) {
              return RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::MOVE_TIMEOUT,
                                 "Timeout, can not complete move to start"};
            }

          } while (semi_auto_state == MotionGeneratorStates::RUNNING_S ||
                   interpreter_state == InterpreterStates::MOTION_NOT_ALLOWED_S ||
                   interpreter_state == InterpreterStates::IN_TRANSITION);

          return motion_not_allowed.first;
        }
        return RobotStatus(mcx_cpp::StatusCode::FAILED, RobotStatus::INTERPRETER_NOT_PLAYING);
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::play(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        req.setParameter(paths.get("INTERPRETER_COMMAND"), InterpreterEvents::PLAY_PROGRAM_E);
        // waits program to start
        auto playing =
            switchToInterpreterState(req, timeout_sec, InterpreterStates::PROGRAM_RUN_S,
                                     {InterpreterStates::MOTION_NOT_ALLOWED_S, InterpreterStates::MSL_PROGRAM_DONE_S,
                                      InterpreterStates::MSL_STATUS_CODE_FAILED},
                                     "Play state", paths);

        return switchToInterpreterState(req, timeout_sec, InterpreterStates::MSL_PROGRAM_DONE_S,
                                        {InterpreterStates::PROGRAM_STOP_S, InterpreterStates::MOTION_NOT_ALLOWED_S,
                                         InterpreterStates::MSL_STATUS_CODE_FAILED},
                                        "Play state", paths)
            .first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::pause(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        req.setParameter(paths.get("INTERPRETER_COMMAND"), InterpreterEvents::PAUSE_PROGRAM_E);
        return switchToInterpreterState(req, timeout_sec, InterpreterStates::PROGRAM_PAUSE_S, {}, "Pause state", paths)
            .first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::stop(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        req.setParameter(paths.get("INTERPRETER_COMMAND"), InterpreterEvents::STOP_PROGRAM_E);
        return switchToInterpreterState(req, timeout_sec, InterpreterStates::PROGRAM_STOP_S, {}, "Stop state", paths)
            .first;
      },
      std::ref(request_), timeout_sec, paths_);
}

CommandReply RobotCommand::reset(double timeout_sec) const {
  return std::async(
      std::launch::async,
      [](const mcx_cpp::Request& req, double timeout_sec, const Paths& paths) {
        req.setParameter(paths.get("INTERPRETER_COMMAND"), InterpreterEvents::RESET_INTERPRETER_E);
        return switchToInterpreterState(req, timeout_sec, InterpreterStates::PROGRAM_STOP_S, {}, "Reset state", paths)
            .first;
      },
      std::ref(request_), timeout_sec, paths_);
}

InterpreterStates RobotCommand::getInterpreterState() const {
  auto result{InterpreterStates::MOTION_NOT_ALLOWED_S};
  request_.getParameter(paths_.get("INTERPRETER_STATE")).get().value(result);
  return result;
}

States RobotCommand::getState() const {
  auto result{States::INIT_S};
  request_.getParameter(paths_.get("LOGIC_STATE")).get().value(result);
  return result;
}

Modes RobotCommand::getMode() const {
  auto result{Modes::INIT_M};
  request_.getParameter(paths_.get("LOGIC_MODE")).get().value(result);
  return result;
}

Paths RobotCommand::getPaths() const { return paths_; }

std::pair<RobotStatus, InterpreterStates> RobotCommand::switchToInterpreterState(
    const mcx_cpp::Request& req, double& timeout_sec, InterpreterStates desired_state,
    std::initializer_list<InterpreterStates> exception_states, const std::string& state_name, const Paths& paths) {
  auto interpreter_state{InterpreterStates::IN_TRANSITION};
  auto logic_state{States::INIT_S};
  do {

    timeout_sec -= measureSec([&] {
      req.getParameter(paths.get("LOGIC_STATE")).get().value(logic_state);
      req.getParameter(paths.get("INTERPRETER_STATE")).get().value(interpreter_state);
    });

    if (logic_state != States::ENGAGED_S) {
      return {RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::NOT_IN_ENGAGE,
                          std::string("System is not in Engage, can not switch interpreter to ").append(state_name)},
              interpreter_state};
    }

    for (auto& exception : exception_states) {
      if (exception == interpreter_state) {
        return {RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::MOTION_NOT_ALLOWED,
                            std::string("Motion is not allowed, can not switch interpreter to ").append(state_name)},
                interpreter_state};
      }
    }
    if (timeout_sec < 0) {
      return {RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::SET_INTERPRETER_STATE_TIMEOUT,
                          std::string("Timeout, can not switch interpreter to ").append(state_name)},
              interpreter_state};
    }
  } while (interpreter_state != desired_state);
  return {RobotStatus{mcx_cpp::StatusCode::OK, RobotStatus::NO_ERROR}, interpreter_state};
}

std::pair<RobotStatus, Modes> RobotCommand::switchToLogicMode(const mcx_cpp::Request& req, double& timeout_sec,
                                                              Modes desired_mode, const std::string& mode_name,
                                                              const Paths& paths) {
  auto robot_mode{Modes::INIT_M};
  auto logic_state{States::ENGAGED_S};
  do {
    timeout_sec -= measureSec([&] {
      // checks if state is still valid
      req.getParameter(paths.get("LOGIC_STATE")).get().value(logic_state);
      // checks actual mode
      req.getParameter(paths.get("LOGIC_MODE")).get().value(robot_mode);
    });
    if (logic_state != States::ENGAGED_S) {
      return {RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::NOT_IN_ENGAGE,
                          std::string("System is not in Engage, can not switch mode to ").append(mode_name)},
              robot_mode};
    }
    if (timeout_sec < 0) {
      return {RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::SET_LOGIC_MODE_TIMEOUT,
                          std::string("Timeout, can not switch logic mode to ").append(mode_name)},
              robot_mode};
    }
  } while (robot_mode != desired_mode);
  return {RobotStatus{mcx_cpp::StatusCode::OK, RobotStatus::NO_ERROR}, robot_mode};
}

std::pair<RobotStatus, States> RobotCommand::switchToLogicState(const mcx_cpp::Request& req, double& timeout_sec,
                                                                States desired_state, const std::string& state_name,
                                                                const Paths& paths) {
  auto robot_state{States::INIT_S};
  do {
    // decrementing timer
    timeout_sec -= measureSec([&] { req.getParameter(paths.get("LOGIC_STATE")).get().value(robot_state); });
    if (robot_state == States::ESTOP_OFF_S) {
      return {RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::ESTOP_ACTIVE,
                          std::string("E-Stop active, can not switch logic state to ").append(state_name)},
              robot_state};
    }
    if (timeout_sec < 0) {
      return {RobotStatus{mcx_cpp::StatusCode::FAILED, RobotStatus::SET_LOGIC_STATE_TIMEOUT,
                          std::string("Timeout, can not switch logic state to ").append(state_name)},
              robot_state};
    }
  } while (robot_state != desired_state);
  return {RobotStatus{mcx_cpp::StatusCode::OK, RobotStatus::NO_ERROR}, robot_state};
}

} // namespace mcx_robot_control