/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#ifndef MOTORCORTEX_CPP_ROBOT_COMMAND_H
#define MOTORCORTEX_CPP_ROBOT_COMMAND_H

#include "paths.h"
#include "system_defs.h"

#include <mcx-cpp/request.h>

namespace mcx_robot_control {

/**@brief Commands to control a state machine of the robot arm.
 * @addtogroup robot command
 * @{
 */

typedef std::future<RobotStatus> CommandReply;

constexpr auto TIMEOUT_INF = std::numeric_limits<double>::max();

class RobotCommand {
public:
  /**@brief Creates new Robot command interface.
   * @param request(motorcortex.Request): reference to a Request instance
   */
  explicit RobotCommand(const mcx_cpp::Request& request);

  /**@brief Acknowledge E-Stop
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply acknowledge(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Switch robot to Off state.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply off(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Switch robot to Disengage state.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply disengage(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Switch robot to Engage state.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply engage(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Switch robot to a Pause mode.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply pauseMode(double timeout_sec) const;

  /**@brief Switch robot to a Manual Cartesian motion.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply manualCartMode(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Switch robot to a Manual Joint motion.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply manualJointMode(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Switch robot to Auto mode (program execution).
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply autoMode(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Switch robot to Semi-auto mode.
   * Semi-auto moves arm to the target when
   * you user holds a button. Semi-auto is
   * active for example during move to start
   * of the program.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply semiAutoMode(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Move arm to a specified pose in joint space.
   * @param angles_rad joint angles in radians
   * @param max_vel maximum joint velocity rad/sec
   * @param max_acc maximum joint acceleration rad/sec²
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply moveToPoint(const std::vector<double>& angles_rad, double max_vel = 0.1, double max_acc = 1.0,
                           double timeout_sec = TIMEOUT_INF) const;

  /**@brief Move arm to the start of the program.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply moveToStart(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Play the program.
   * @param timeout_sec execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply play(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Pause the program.
   * @param command execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply pause(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Stop the program.
   * @param command execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply stop(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Stop the program and clear the interpreter buffer.
   * @param command execution timeout in seconds
   * @return status of the command execution.
   */
  CommandReply reset(double timeout_sec = TIMEOUT_INF) const;

  /**@brief Returns actual state of the interpreter.
   * @return actual state of the interpreter.
   */
  [[nodiscard]] InterpreterStates getInterpreterState() const;

  /**@brief Returns actual state of the robot.
   * @return actual state of the robot.
   */
  [[nodiscard]] States getState() const;

  /**@brief Returns actual mode of the robot.
   * @return actual mode of the robot.
   */
  [[nodiscard]] Modes getMode() const;

  /**@brief Returns system paths, which are used to control the robot
   * @return system paths.
   */
  [[nodiscard]] Paths getPaths() const;

  /** @} End of Doxygen Groups*/

private:
  static std::pair<RobotStatus, InterpreterStates>
  switchToInterpreterState(const mcx_cpp::Request& req, double& timeout_sec, InterpreterStates desired_state,
                           std::initializer_list<InterpreterStates> exception_states, const std::string& state_name,
                           const Paths& paths);

  static std::pair<RobotStatus, Modes> switchToLogicMode(const mcx_cpp::Request& req, double& timeout_sec,
                                                         Modes desired_mode, const std::string& mode_name,
                                                         const Paths& paths);

  static std::pair<RobotStatus, States> switchToLogicState(const mcx_cpp::Request& req, double& timeout_sec,
                                                           States desired_state, const std::string& state_name,
                                                           const Paths& paths);

  const mcx_cpp::Request& request_;

  Paths paths_;
};

} // namespace mcx_robot_control

#endif // MOTORCORTEX_CPP_ROBOT_COMMAND_H
