/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#ifndef MOTORCORTEX_CPP_WAYPOINT_H
#define MOTORCORTEX_CPP_WAYPOINT_H

#include "motionSL.pb.h"
#include <utility>
#include <vector>

namespace mcx_robot_control {

/**@brief Class represents a waypoint of the motion path
 * @addtogroup motion program
 * @{
 */

class Waypoint {
public:
  /**@brief Creates new trajectory waypoint with constraints
   * @param pose: pose in Cartesian or joint space
   * @param smoothing_factor: waypoint smoothing factor in the range [0..1]
   * @param segment_velocity: segment velocity factor in the range [0..1]
   */
  explicit Waypoint(std::vector<double> pose, double smoothing_factor = 0.1, double segment_velocity = 1.0)
      : coord_{std::move(pose)},
        constraint_{.type = motion_spec_WAYPOINTCONSTRAINTTYPE_TIME, .factor = smoothing_factor},
        segment_velocity_{segment_velocity} {}

  /**@brief Returns a coordinate of the waypoint pose
   * @param index: index of the requested coordinate
   * @return requested coordinate
   */
  [[nodiscard]] double coordinate(size_t index) const { return index < coord_.size() ? coord_[index] : 0; }

  /**@brief Returns a length/dimension of the coordinate vector
   * @return a dimension of the coordinate vector
   */
  [[nodiscard]] size_t length() const { return coord_.size(); }

  /**@brief Returns a smooth factor
   * @return smooth factor
   */
  [[nodiscard]] double smoothFactor() const { return constraint_.factor; }

  /**@brief Returns a type of the waypoint constraint
   * @return a type of the waypoint constraint
   */
  [[nodiscard]] motion_spec_WAYPOINTCONSTRAINTTYPE type() const { return constraint_.type; }

  /**@brief Returns a segment velocity factor
   * @return a segment velocity factor
   */
  [[nodiscard]] double segmentVelocity() const { return segment_velocity_; }

  /** @} End of Doxygen Groups*/

private:
  std::vector<double> coord_;
  motion_spec_WayPointConstraint constraint_{};
  double segment_velocity_{};
};

} // namespace mcx_robot_control

#endif // MOTORCORTEX_CPP_WAYPOINT_H
