/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#include "helper.h"
#include "mcx-robot-control.h"
#include <iostream>

using namespace std;
using namespace mcx_robot_control;

constexpr double TIMEOUT_SEC = 5;

void moveToStart(const RobotCommand& robot) {
  clog << "Move To Start Command" << endl;
  auto move_done = robot.moveToStart().get();
  if (move_done.status() == mcx_cpp::StatusCode::OK) {
    clog << "Interpreter is in Move to Start State" << endl;
  } else {
    helper::printStatus(move_done);
    throw(runtime_error(string("Failed to set interpreter to Move to Start State")));
  }
}

bool playState(const RobotCommand& robot) {
  clog << "Play Command" << endl;
  auto play_done = robot.play().get();
  if (play_done.status() == mcx_cpp::StatusCode::OK) {
    clog << "Interpreter played program" << endl;
    return true;
  } else if (play_done.error() == RobotStatus::ErrorCode::MOTION_NOT_ALLOWED) {
    clog << "Can not play program, move robot to the start position" << endl;
    return false;
  } else {
    helper::printStatus(play_done);
    throw(runtime_error(string("Failed to set interpreter to Play State")));
  }
}

void stopState(const RobotCommand& robot) {
  auto stop_done = robot.stop(TIMEOUT_SEC).get();
  if (stop_done.status() == mcx_cpp::StatusCode::OK) {
    clog << "Interpreter is in Stop State" << endl;
  } else {
    helper::printStatus(stop_done);
    throw(runtime_error(string("Failed to set interpreter to Stop State")));
  }
}

void pauseState(const RobotCommand& robot) {
  auto pause_done = robot.pause(TIMEOUT_SEC).get();
  if (pause_done.status() == mcx_cpp::StatusCode::OK) {
    clog << "Interpreter is in Pause State" << endl;
  } else {
    helper::printStatus(pause_done);
    throw(runtime_error(string("Failed to set interpreter to Pause State")));
  }
}

void resetState(const RobotCommand& robot) {}

int main(int argc, char* argv[]) {

  clog << "Motorcortex Robot Control library version: " << mcx_cpp::version() << endl;

  // setting url
  // url format: ws://address:req_port:sub_port
  //  std::string url = "wss://localhost:5568:5567";
  std::string url = "wss://192.168.2.100:5568:5567";
  mcx_cpp::ConnectionOptions options{"mcx.cert.pem"};
  if (argc > 1) {
    url = std::string(argv[1]);
  }

  // Creating a parameter tree;
  mcx_cpp::ParameterTree parameter_tree;
  mcx_cpp::Request req{parameter_tree};
  mcx_cpp::Subscribe sub{req};
  helper::connect(url, options, parameter_tree, req, sub);

  // Subscribing to the log
  sub.subscribe({"root/logger/logOut"}, "log", 1).get().notify([](auto params) {
    std::string log;
    params[0].value(log);
    clog << "Log: " << log;
  });

  RobotCommand robot{req};
  robot.acknowledge();
  robot.engage();
  robot.reset();

  // building motion program
  MotionProgram motion_program{req};
  auto start_position_jnt = Waypoint(toRadians({0.0, 0.0, 90.0, 0.0, 90.0, 0.0}));
  auto jpos_1 = Waypoint(toRadians({15.0, 0.0, 90.0, 0.0, 90.0, 0.0}));
  auto jpos_2 = Waypoint(toRadians({30.0, 15.0, 90.0, 0.0, 90.0, 0.0}));
  auto jpos_3 = Waypoint(toRadians({-30.0, -15.0, 90.0, 15.0, 90.0, 15.0}));

  motion_program.addMoveJ(start_position_jnt, 0.5, 0.5);
  motion_program.addMoveJ({jpos_1, jpos_2}, 1.0, 0.5);
  motion_program.addMoveJ({jpos_2, jpos_3}, 0.5, 1.0);
  motion_program.addMoveJ(start_position_jnt, 1.0, 0.3);

  // play program
  for (size_t i = 0; i < 5; i++) {
    auto program_sent = motion_program.send("example1", static_cast<int>(i)).get();
    if (program_sent.status() == mcx_cpp::StatusCode::OK) {
      clog << "Motion Program sent" << endl;
    } else {
      helper::printStatus(program_sent);
      throw(runtime_error(string("Failed to send Motion Program")));
    }
    if (!playState(robot)) {
      moveToStart(robot);
      playState(robot);
    }
  }

  for (size_t i = 0; i < 5; i++) {
    stopState(robot);
    if (!playState(robot)) {
      moveToStart(robot);
      playState(robot);
    }
  }

  robot.off();
}