/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#include "helper.h"
#include "mcx-robot-control.h"

using namespace std;
using namespace mcx_robot_control;

int main(int argc, char* argv[]) {

  clog << "Motorcortex Robot Control library version: " << mcx_cpp::version() << endl;

  // setting url
  // url format: ws://address:req_port:sub_port
  std::string url = "wss://localhost:5568:5567";
  mcx_cpp::ConnectionOptions options{"mcx.cert.pem"};
  if (argc > 1) {
    url = std::string(argv[1]);
  }

  // Creating a parameter tree;
  mcx_cpp::ParameterTree parameter_tree;
  mcx_cpp::Request req{parameter_tree};
  mcx_cpp::Subscribe sub{req};
  helper::connect(url, options, parameter_tree, req, sub);

  // Subscribing to the log
  sub.subscribe({"root/logger/logOut"}, "log", 1).get().notify([](auto params) {
    std::string log;
    params[0].value(log);
    clog << "Log: " << log;
  });

  RobotCommand robot{req};
  robot.acknowledge();
  robot.engage();
  robot.reset();

  // building motion program
  MotionProgram motion_program{req};
  auto start_position_jnt = Waypoint(toRadians({0.0, 0.0, 90.0, 0.0, 90.0, 0.0}));
  auto jpos_1 = Waypoint(toRadians({15.0, 0.0, 90.0, 0.0, 90.0, 0.0}));
  auto jpos_2 = Waypoint(toRadians({30.0, 15.0, 90.0, 0.0, 90.0, 0.0}));
  auto jpos_3 = Waypoint(toRadians({-30.0, -15.0, 90.0, 15.0, 90.0, 15.0}));

  motion_program.addMoveJ(start_position_jnt, 0.5, 0.5);
  motion_program.addMoveJ({jpos_1, jpos_2}, 1.0, 0.5);
  motion_program.addMoveJ({jpos_2, jpos_3}, 0.5, 1.0);
  motion_program.addMoveJ(start_position_jnt, 1.0, 0.3);

  auto program_sent = motion_program.send("example1", 0).get();
  if (program_sent.status() == mcx_cpp::StatusCode::OK) {
    clog << "Motion Program sent" << endl;
  } else {
    helper::printStatus(program_sent);
    throw(runtime_error(string("Failed to send Motion Program")));
  }

  // playing motion program
  robot.play();
  robot.moveToStart();
  for (int i = 0; i < 5; i++) {
    motion_program.send("example1", i);
    auto playing = robot.play().get();
    if (playing.status() == mcx_cpp::StatusCode::OK) {
      clog << "Motion Program done, id: " << i << endl;
    } else {
      helper::printStatus(playing);
      throw(runtime_error(string("Failed to play Motion Program id: ").append(std::to_string(i))));
    }
  }

  robot.off();
}