/*
 * Developer : Alexey Zakharov (alexey.zakharov@vectioneer.com)
 * All rights reserved. Copyright (c) 2019 VECTIONEER.
 */

#include "helper.h"
#include "mcx-robot-control.h"

using namespace std;
using namespace mcx_robot_control;

int main(int argc, char* argv[]) {

  clog << "Motorcortex Robot Control library version: " << mcx_cpp::version() << endl;

  // setting url
  //   url format: ws://address:req_port:sub_port
  std::string url = "wss://localhost:5568:5567";
  // std::string url = "wss://192.168.2.100:5568:5567";
  mcx_cpp::ConnectionOptions options{"mcx.cert.pem"};
  if (argc > 1) {
    url = std::string(argv[1]);
  }

  // Creating a parameter tree;
  mcx_cpp::ParameterTree parameter_tree;
  mcx_cpp::Request req{parameter_tree};
  mcx_cpp::Subscribe sub{req};
  helper::connect(url, options, parameter_tree, req, sub);

  // Subscribing to the log
  sub.subscribe({"root/logger/logOut"}, "log", 1).get().notify([](auto params) {
    std::string log;
    params[0].value(log);
    clog << "Log: " << log;
  });

  RobotCommand robot{req};
  robot.acknowledge();
  auto engage_status = robot.engage().get();
  if (engage_status.status() != mcx_cpp::StatusCode::OK) {
    helper::printStatus(engage_status);
  }
  robot.reset();

  // building motion program
  MotionProgram motion_program{req};
  auto start_position_cart = Waypoint({0.4145, 0.0, 0.334, 3.14159, 0.0, 3.14159});
  auto cart_1 = Waypoint({0.400376, 0.10728, 0.334, -2.87979, 0.0, 3.14159});
  auto cart_2 = Waypoint({0.392097, 0.226377, 0.239572, -2.61799, -0.261799, 3.14159});
  auto cart_3 = Waypoint({0.325589, -0.139415, 0.420658, 2.36076, 0.186372, 2.82229});

  static constexpr auto DUMMY_BOOL_FLAG = "root/UserParameters/IO/Var";
  if (!parameter_tree.has(DUMMY_BOOL_FLAG)) {
    throw(runtime_error(string("Failed to find parameter: ") + DUMMY_BOOL_FLAG));
  }

  motion_program.addMoveL(start_position_cart, 0.5, 0.5);
  motion_program.addWait(1);
  motion_program.addMoveL({cart_1, cart_2}, 1.0, 0.5);
  motion_program.addWait(1);
  motion_program.addMoveL({cart_2, cart_3}, 0.5, 0.5);
  motion_program.addWait(DUMMY_BOOL_FLAG, true, 500);
  motion_program.addMoveL(start_position_cart, 1.0, 1.0);

  future<void> async_command = async(
      [](const mcx_cpp::Request& req) {
        this_thread::sleep_for(std::chrono::seconds(20));
        auto status = req.setParameter(DUMMY_BOOL_FLAG, true).get();
        if (status.status() == mcx_cpp::StatusCode::OK) {
          clog << "Wait flag is set to True, continue motion" << endl;
        } else {
          helper::printStatus(status);
          throw(runtime_error(string("Failed to set wait flag")));
        }
      },
      std::ref(req));

  req.setParameter(DUMMY_BOOL_FLAG, false);
  auto program_sent = motion_program.send("example1", 0).get();
  if (program_sent.status() == mcx_cpp::StatusCode::OK) {
    clog << "Motion Program sent" << endl;
  } else {
    helper::printStatus(program_sent);
    throw(runtime_error(string("Failed to send Motion Program")));
  }

  // playing motion program
  robot.play();
  robot.moveToStart();
  for (int i = 0; i < 5; i++) {
    motion_program.send("example1", i);
    auto playing = robot.play().get();
    if (playing.status() == mcx_cpp::StatusCode::OK) {
      clog << "Motion Program done, id: " << i << endl;
    } else {
      helper::printStatus(playing);
      throw(runtime_error(string("Failed to play Motion Program id: ").append(std::to_string(i))));
    }
  }

  robot.off();
}